import { BaseManager } from "./baseManager"

export class BaseCRUDManager extends BaseManager {
  constructor(modelName, type, container) {
    super(modelName, type, container)
    this.mongoDb = container.graphqlMongoDb
    this.pgsqlDb = container.graphqlPGSqlDb
  }

  /** Override default for other db system */
  getDb() {
    return this.mongoDb
  }

  // ============== Standard CRUD Methods ======================

  async get(args, { auth }, { doAuth, session }) {
    const api = this.getDb().getAPI(this.modelName)
    const data = await api.get({ id: args.id, session: session })
    await this.authorize("get", doAuth, data, auth)
    const result = await this.applyReadPermissions(doAuth, data, auth)
    return result
  }

  async query(args, { auth }, { doAuth, session }) {
    await this.authorize("query", doAuth, null, auth)
    const results = await this.getDb()
      .getAPI(this.modelName)
      .query(args.offset, args.limit, args.filter, args.sort, session)

    results.items = await this.applyQueryReadPermissions(
      doAuth,
      results.items,
      auth
    )

    return results
  }

  async create(data, { auth }, { doAuth, session }) {
    await this.authorize("create", doAuth, null, auth)
    const api = this.getDb().getAPI(this.modelName)
    const result = await api.create(data, session)
    return {
      success: true,
      message: `${this.modelName} Created`,
      data: result,
    }
  }

  async update(id, data, { auth }, { doAuth, session }) {
    const api = this.getDb().getAPI(this.modelName)
    const obj = await api.get({ id: id }) // fetch to do authorization
    await this.authorize("update", doAuth, obj, auth)

    const result = await api.update(id, data, session)
    return {
      success: true,
      message: `${this.modelName} Updated`,
      data: result,
    }
  }

  async delete(id, { auth }, { doAuth, session }) {
    const api = this.getDb().getAPI(this.modelName)
    const obj = await api.get({ id: id }) // fetch to do authorization
    await this.authorize("delete", doAuth, obj, auth)

    let result = await api.delete(id, session)

    const props = Object.getOwnPropertyNames(result)

    if (props.length == 0) {
      return {
        success: false,
        message: `Delete Failed`,
        data: null,
      }
    }

    return {
      success: true,
      message: `${this.modelName} Deleted`,
      data: result,
    }
  }

  // ===================================================
  // Helper methods
  // Many-to-one fetch from another Type
  async getLinked({ obj, byField }, context, options) {
    let result = null

    if (byField in obj && obj[byField] != null) {
      const linkId = obj[byField]
      try {
        result = await this.get({ id: linkId }, context, options)
      } catch (ex) {
        this.log.error(
          `BaseCRUDManager.getLinked Error getting linked object: ${
            ex.constructor.name
          } msg: ${ex.message}`
        )
      }
    } else {
      // this.log.warn(
      //   `getLinked byField not found: ${byField} ${JSON.stringify(
      //     obj,
      //     null,
      //     2
      //   )}`
      // )
    }

    return result
  }

  // ===================================================
  // Get list of related objects based on this.id == foreign.
  async getLinkedMulti(
    { obj, idField, byForeignField, filter, sort, offset, limit },
    context,
    options
  ) {
    idField = idField || "id"
    let result = null

    if (idField in obj && obj[idField] != null) {
      const keyValue = obj[idField]
      let keyFilter = `${byForeignField} = '${keyValue}'`
      if (filter) {
        keyFilter = `(${keyFilter}) AND (${filter})`
      }
      try {
        result = await this.query(
          { filter: keyFilter, sort, offset, limit },
          context,
          options
        )
      } catch (ex) {
        this.log.error(
          `BaseCRUDManager.getLinked Error getting linked object: ${
            ex.constructor.name
          } msg: ${ex.message}`
        )
      }
    }

    return result ? result.items : null
  }
}
