import path from "path"
import fs from "fs"
import { merge, concat } from "lodash"
import { isFunction } from "util"
import { gql } from "apollo-server"

const manifestName = "Manifest"
const queriesSuffix = "Queries"
const resolversSuffux = "Resolvers"
const managerSuffix = "Manager"

const emptyQuery = gql`
  type Query {
    _empty: String
  }

  type Mutation {
    _empty: String
  }
`

export class TypesManager {
  constructor(container) {
    this.container = container
    this.log = container.log
    this.types = {}
    this.schemas = []
    this.queries = [emptyQuery]
    this.resolvers = {}
    this.managers = {}
  }

  async loadTypeModules(basePath) {
    return new Promise((resolve, reject) => {
      fs.readdir(basePath, (err, items) => {
        if (err) {
          reject(err)
        } else {
          items.forEach(item => {
            const itemPath = path.join(basePath, item)
            const isDir = fs.lstatSync(itemPath).isDirectory()
            if (isDir) {
              try {
                const typeModule = require(itemPath)
                let typeModuleItems = Object.keys(typeModule)
                if (typeModuleItems.includes(manifestName)) {
                  const manifest = typeModule.Manifest
                  typeModuleItems = this.arrayRemove(
                    typeModuleItems,
                    manifestName
                  )

                  const types = manifest.types || []
                  let baseTypeName = this.firstUpperCase(item)
                  if (types.length > 0) {
                    baseTypeName = types[0]
                  }

                  //load Queries
                  typeModuleItems = this.loadQueries(
                    baseTypeName,
                    manifest,
                    typeModule,
                    typeModuleItems
                  )
                  // load Resolvers
                  typeModuleItems = this.loadResolvers(
                    baseTypeName,
                    manifest,
                    typeModule,
                    typeModuleItems
                  )
                  // load Manager
                  typeModuleItems = this.loadManager(
                    baseTypeName,
                    manifest,
                    typeModule,
                    typeModuleItems
                  )
                  // load all Types
                  this.loadTypes(
                    baseTypeName,
                    manifest,
                    typeModule,
                    typeModuleItems
                  )
                } else {
                  this.log.warn(`No Manifest: "${item}" Is not a legit type.`)
                }
              } catch (ex) {
                this.log.warn(`${item}: ${ex}`)
              }
            }
          })
        }

        resolve(true)
      })
    })
  }

  loadQueries(baseTypeName, manifest, typeModule, typeModuleItems) {
    let queriesName = manifest.queries || baseTypeName + queriesSuffix
    typeModuleItems = this.arrayRemove(typeModuleItems, queriesName)
    const queries = typeModule[queriesName] || null
    if (queries) {
      this.queries = this.queries.concat(queries)
    } else {
      this.log.warn(`No Queries Found for ${baseTypeName}`)
    }
    return typeModuleItems
  }

  loadResolvers(baseTypeName, manifest, typeModule, typeModuleItems) {
    let resolversName = manifest.resolvers || baseTypeName + resolversSuffux
    typeModuleItems = this.arrayRemove(typeModuleItems, resolversName)
    const resolvers = typeModule[resolversName] || null
    if (resolvers) {
      this.resolvers = merge(this.resolvers, resolvers)
    } else {
      this.log.warn(`No Resolvers Found for ${baseTypeName}`)
    }
    return typeModuleItems
  }

  loadManager(baseTypeName, manifest, typeModule, typeModuleItems) {
    const typeClass = typeModule[baseTypeName] || null
    if (!typeClass) {
      throw new Error(`No Type found for : ${baseTypeName}`)
    }
    let managerName = manifest.manager || baseTypeName + managerSuffix
    typeModuleItems = this.arrayRemove(typeModuleItems, managerName)
    const managerClass = typeModule[managerName] || null
    if (managerClass) {
      const manager = new managerClass(baseTypeName, typeClass, this.container)
      this.managers[baseTypeName] = manager
    } else {
      this.log.warn(`No Manager found for ${baseTypeName}`)
    }
    return typeModuleItems
  }

  /**
   * If the manifest declares the types, use that list.
   * If it does not, the list in typeModuleItems should contain whatever is left over
   * after all other items are loaded and redacted from the list.
   * As such it should tolerate erroneous entries with warnings.
   * @param {*} baseTypeName
   * @param {*} manifest
   * @param {*} typeModule
   * @param {*} typeModuleItems
   */
  loadTypes(baseTypeName, manifest, typeModule, typeModuleItems) {
    const loadTypeNames = manifest.types || typeModuleItems
    loadTypeNames.forEach(typeName => {
      this.loadType(typeName, typeModule)
    })
  }

  loadType(typeName, typeModule) {
    const typeClass = typeModule[typeName] || null
    if (typeClass && isFunction(typeClass)) {
      try {
        const typeObj = new typeClass()
        this.types[typeName] = typeObj
        const schema = typeObj.toGraphQLSchema()
        this.schemas = this.schemas.concat(schema)
      } catch (ex) {
        this.log.warn(`Error loading Txype: ${ex}`)
      }
    } else {
      this.log.error(`loadType not a function: ${typeName}`)
    }
  }

  arrayRemove(arr, itemName) {
    const ix = arr.indexOf(itemName)
    if (ix >= 0) {
      arr = concat(arr.slice(0, ix), arr.slice(ix + 1))
    }
    return arr
  }

  firstUpperCase(str) {
    if (typeof str === "string") {
      return str[0].toUpperCase() + str.slice(1)
    }
    return null
  }

  getTypeDefs() {
    return this.queries.concat(this.schemas)
  }

  getResolvers() {
    return this.resolvers
  }

  getManagers() {
    return this.managers
  }
}
